@@ -21,13 +21,16 @@
 #include <stdlib.h>
 #include <argp.h>
 #include <string.h>
-#ifdef HAVE_LIBKRUN
-#  include <libgen.h>
+#include <libgen.h>
+
+#ifdef HAVE_DLOPEN
+#  include <dlfcn.h>
 #endif
 
 #include "crun.h"
 #include "libcrun/utils.h"
 #include "libcrun/custom-handler.h"
+#include "libcrun/status.h"
 
 /* Commands.  */
 #include "run.h"
@@ -95,6 +98,8 @@ init_libcrun_context (libcrun_context_t *con, const char *id, struct crun_global
   con->force_no_cgroup = glob->option_force_no_cgroup;
   con->notify_socket = getenv ("NOTIFY_SOCKET");
   con->fifo_exec_wait_fd = -1;
+  con->argc = glob->argc;
+  con->argv = glob->argv;
 
   ret = libcrun_init_logging (&con->output_handler, &con->output_handler_arg, id, glob->log, err);
   if (UNLIKELY (ret < 0))
@@ -147,7 +152,7 @@ struct commands_s commands[] = { { COMMAND_CREATE, "create", crun_command_create
                                  { COMMAND_UPDATE, "update", crun_command_update },
                                  { COMMAND_PAUSE, "pause", crun_command_pause },
                                  { COMMAND_UNPAUSE, "resume", crun_command_unpause },
-#ifdef HAVE_CRIU
+#if HAVE_CRIU && HAVE_DLOPEN
                                  { COMMAND_CHECKPOINT, "checkpoint", crun_command_checkpoint },
                                  { COMMAND_RESTORE, "restore", crun_command_restore },
 #endif
@@ -156,7 +161,7 @@ struct commands_s commands[] = { { COMMAND_CREATE, "create", crun_command_create
                                  } };
 
 static char doc[] = "\nCOMMANDS:\n"
-#ifdef HAVE_CRIU
+#if HAVE_CRIU && HAVE_DLOPEN
                     "\tcheckpoint  - checkpoint a container\n"
 #endif
                     "\tcreate      - create a container\n"
@@ -165,7 +170,7 @@ static char doc[] = "\nCOMMANDS:\n"
                     "\tlist        - list known containers\n"
                     "\tkill        - send a signal to the container init process\n"
                     "\tps          - show the processes in the container\n"
-#ifdef HAVE_CRIU
+#if HAVE_CRIU && HAVE_DLOPEN
                     "\trestore     - restore a container\n"
 #endif
                     "\trun         - run a container\n"
@@ -190,6 +195,8 @@ get_command (const char *arg)
 
 enum
 {
+  OPTION_VERSION = 'v',
+  OPTION_VERSION_CAP = 'V',
   OPTION_DEBUG = 1000,
   OPTION_SYSTEMD_CGROUP,
   OPTION_CGROUP_MANAGER,
@@ -199,7 +206,6 @@ enum
   OPTION_ROOTLESS
 };
 
-const char *argp_program_version = PACKAGE_STRING;
 const char *argp_program_bug_address = "https://github.com/containers/crun/issues";
 
 static struct argp_option options[] = { { "debug", OPTION_DEBUG, 0, 0, "produce verbose output", 0 },
@@ -209,6 +215,9 @@ static struct argp_option options[] = { { "debug", OPTION_DEBUG, 0, 0, "produce
                                         { "log-format", OPTION_LOG_FORMAT, "FORMAT", 0, NULL, 0 },
                                         { "root", OPTION_ROOT, "DIR", 0, NULL, 0 },
                                         { "rootless", OPTION_ROOT, "VALUE", 0, NULL, 0 },
+                                        { "version", OPTION_VERSION, 0, 0, NULL, 0 },
+                                        // alias OPTION_VERSION_CAP with OPTION_VERSION
+                                        { NULL, OPTION_VERSION_CAP, 0, OPTION_ALIAS, NULL, 0 },
                                         {
                                             0,
                                         } };
@@ -216,8 +225,10 @@ static struct argp_option options[] = { { "debug", OPTION_DEBUG, 0, 0, "produce
 static void
 print_version (FILE *stream, struct argp_state *state arg_unused)
 {
+  cleanup_free char *rundir = libcrun_get_state_directory (arguments.root, NULL);
   fprintf (stream, "%s version %s\n", PACKAGE_NAME, PACKAGE_VERSION);
   fprintf (stream, "commit: %s\n", GIT_VERSION);
+  fprintf (stream, "rundir: %s\n", rundir);
   fprintf (stream, "spec: 1.0.0\n");
 #ifdef HAVE_SYSTEMD
   fprintf (stream, "+SYSTEMD ");
@@ -300,6 +311,10 @@ parse_opt (int key, char *arg, struct argp_state *state)
     case ARGP_KEY_NO_ARGS:
       libcrun_fail_with_error (0, "please specify a command");
 
+    case OPTION_VERSION:
+    case OPTION_VERSION_CAP:
+      print_version (stdout, state);
+      exit (EXIT_SUCCESS);
     default:
       return ARGP_ERR_UNKNOWN;
     }
@@ -328,20 +343,45 @@ argp_mandatory_argument (char *arg, struct argp_state *state)
 
 static struct argp argp = { options, parse_opt, args_doc, doc, NULL, NULL, NULL };
 
+int ensure_cloned_binary (void);
+
+static void
+fill_handler_from_argv0 (char *argv0, struct crun_global_arguments *args)
+{
+  const char *b = basename (argv0);
+#ifdef HAVE_LIBKRUN
+  if (strcmp (b, "krun") == 0)
+    {
+      args->handler = "krun";
+      return;
+    }
+#endif
+  if (has_prefix (b, "crun-") && b[5] != '\0')
+    args->handler = b + 5;
+}
+
 int
 main (int argc, char **argv)
 {
   libcrun_error_t err = NULL;
   int ret, first_argument = 0;
 
-  argp_program_version_hook = print_version;
-#ifdef HAVE_LIBKRUN
-  if (strcmp (basename (argv[0]), "krun") == 0)
+  arguments.argc = argc;
+  arguments.argv = argv;
+
+#ifdef DYNLOAD_LIBCRUN
+  if (ensure_cloned_binary () < 0)
     {
-      arguments.handler = "krun";
+      fprintf (stderr, "Failed to re-execute libcrun via memory file descriptor\n");
+      _exit (EXIT_FAILURE);
     }
+  /* Resolve all libcrun weak dependencies.  */
+  if (dlopen ("libcrun.so", RTLD_GLOBAL | RTLD_DEEPBIND | RTLD_LAZY) == NULL)
+    error (EXIT_FAILURE, 0, "dlopen: %s", dlerror ());
 #endif
 
+  fill_handler_from_argv0 (argv[0], &arguments);
+
   argp_parse (&argp, argc, argv, ARGP_IN_ORDER, &first_argument, &arguments);
 
   command = get_command (argv[first_argument]);
